TEX_TO_SYMBOL_COUNT = {
    R"\!": 0,
    R"\,": 0,
    R"\-": 0,
    R"\/": 0,
    R"\:": 0,
    R"\;": 0,
    R"\>": 0,
    R"\aa": 0,
    R"\AA": 0,
    R"\ae": 0,
    R"\AE": 0,
    R"\arccos": 6,
    R"\arcsin": 6,
    R"\arctan": 6,
    R"\arg": 3,
    R"\author": 0,
    R"\bf": 0,
    R"\bibliography": 0,
    R"\bibliographystyle": 0,
    R"\big": 0,
    R"\Big": 0,
    R"\bigodot": 4,
    R"\bigoplus": 5,
    R"\bigskip": 0,
    R"\bmod": 3,
    R"\boldmath": 0,
    R"\bottomfraction": 2,
    R"\bowtie": 2,
    R"\cal": 0,
    R"\cdots": 3,
    R"\centering": 0,
    R"\cite": 2,
    R"\cong": 2,
    R"\contentsline": 0,
    R"\cos": 3,
    R"\cosh": 4,
    R"\cot": 3,
    R"\coth": 4,
    R"\csc": 3,
    R"\date": 0,
    R"\dblfloatpagefraction": 2,
    R"\dbltopfraction": 2,
    R"\ddots": 3,
    R"\deg": 3,
    R"\det": 3,
    R"\dim": 3,
    R"\displaystyle": 0,
    R"\div": 2,
    R"\doteq": 2,
    R"\dotfill": 0,
    R"\dots": 3,
    R"\emph": 0,
    R"\exp": 3,
    R"\fbox": 4,
    R"\floatpagefraction": 2,
    R"\flushbottom": 0,
    R"\footnotesize": 0,
    R"\footnotetext": 0,
    R"\frame": 2,
    R"\framebox": 4,
    R"\fussy": 0,
    R"\gcd": 3,
    R"\ghost": 0,
    R"\glossary": 0,
    R"\hfill": 0,
    R"\hom": 3,
    R"\hookleftarrow": 2,
    R"\hookrightarrow": 2,
    R"\hrulefill": 0,
    R"\huge": 0,
    R"\Huge": 0,
    R"\hyphenation": 0,
    R"\iff": 2,
    R"\Im": 2,
    R"\index": 0,
    R"\inf": 3,
    R"\it": 0,
    R"\ker": 3,
    R"\l": 0,
    R"\L": 0,
    R"\label": 0,
    R"\large": 0,
    R"\Large": 0,
    R"\LARGE": 0,
    R"\ldots": 3,
    R"\lefteqn": 0,
    R"\left": 0,
    R"\lg": 2,
    R"\lim": 3,
    R"\liminf": 6,
    R"\limsup": 6,
    R"\linebreak": 0,
    R"\ln": 2,
    R"\log": 3,
    R"\longleftarrow": 2,
    R"\Longleftarrow": 2,
    R"\longleftrightarrow": 2,
    R"\Longleftrightarrow": 2,
    R"\longmapsto": 3,
    R"\longrightarrow": 2,
    R"\Longrightarrow": 2,
    R"\makebox": 0,
    R"\mapsto": 2,
    R"\markright": 0,
    R"\mathds": 0,
    R"\max": 3,
    R"\mbox": 0,
    R"\medskip": 0,
    R"\min": 3,
    R"\mit": 0,
    R"\models": 2,
    R"\ne": 2,
    R"\neq": 2,
    R"\newline": 0,
    R"\noindent": 0,
    R"\nolinebreak": 0,
    R"\nonumber": 0,
    R"\nopagebreak": 0,
    R"\normalmarginpar": 0,
    R"\normalsize": 0,
    R"\notin": 2,
    R"\o": 0,
    R"\O": 0,
    R"\obeycr": 0,
    R"\oe": 0,
    R"\OE": 0,
    R"\overbrace": 4,
    R"\pagebreak": 0,
    R"\pagenumbering": 0,
    R"\pageref": 2,
    R"\pmod": 5,
    R"\Pr": 2,
    R"\protect": 0,
    R"\qquad": 0,
    R"\quad": 0,
    R"\raggedbottom": 0,
    R"\raggedleft": 0,
    R"\raggedright": 0,
    R"\Re": 2,
    R"\ref": 2,
    R"\restorecr": 0,
    R"\reversemarginpar": 0,
    R"\right": 0,
    R"\rm": 0,
    R"\sc": 0,
    R"\scriptscriptstyle": 0,
    R"\scriptsize": 0,
    R"\scriptstyle": 0,
    R"\sec": 3,
    R"\sf": 0,
    R"\shortstack": 0,
    R"\sin": 3,
    R"\sinh": 4,
    R"\sl": 0,
    R"\sloppy": 0,
    R"\small": 0,
    R"\Small": 0,
    R"\smallskip": 0,
    R"\sqrt": 2,
    R"\ss": 0,
    R"\sup": 3,
    R"\tan": 3,
    R"\tanh": 4,
    R"\text": 0,
    R"\textbf": 0,
    R"\textfraction": 2,
    R"\textstyle": 0,
    R"\thicklines": 0,
    R"\thinlines": 0,
    R"\thinspace": 0,
    R"\tiny": 0,
    R"\title": 0,
    R"\today": 15,
    R"\topfraction": 2,
    R"\tt": 0,
    R"\typeout": 0,
    R"\unboldmath": 0,
    R"\underbrace": 6,
    R"\underline": 0,
    R"\value": 0,
    R"\vdots": 3,
    R"\vline": 0
}