from manimx.event_handler.event_dispatcher import EventDispatcher


# This is supposed to be a Singleton
# i.e., during runtime there should be only one object of Event Dispatcher
EVENT_DISPATCHER = EventDispatcher()
